; #########################################################################
;
;                             ZAtanh
;
;##########################################################################

  ; -----------------------------------------------------------------------
  ; This procedure was written by Raymond Filiatreault, February 2005
  ; 
  ;            atanh(Src) -> Dest  = 1/2 * ln((1+z) / (1-z))
  ;
  ; This ZAtanh function computes the hyperbolic arctangent of the complex
  ; number Src1 and returns the result as a complex number at the specified
  ; destination (the FPU itself or a memory location), unless an invalid
  ; operation is detected or the definition of the parameters (with uID)
  ; is invalid.
  ;
  ; The source can be either in cartesian or polar coordinates. The result
  ; will be returned in the same coordinates.
  ;
  ; The source can be a complex number already on the FPU or from memory. If
  ; on the FPU, the REAL coefficient must be in the ST(0) register and the
  ; IMAGINARY coefficient in the ST(1) register. Memory variables must have
  ; the two coefficients at adjacent memory addresses, the REAL coefficient
  ; being at the lower address. All memory sources and destinations must
  ; have the same declared size as specified in the uID parameter.
  ;
  ; The source is not checked for validity. This is the programmer's
  ; responsibility. The programmer is also responsible to insure that all
  ; the FPU data registers necessary for performing the function are FREE.
  ;  If either Src1 or Src2 is on the FPU:  ALL other registers must be free
  ;  Otherwise: ALL registers must be free
  ;
  ; Only EAX is used to return error (0) or success (!0). All other CPU
  ; registers are preserved.
  ;
  ; If a valid operand size is present in the uID, any source data on the
  ; FPU will be modified; otherwise, if an invalid operand size is present
  ; in the uID, all FPU registers are preserved and the location of the
  ; destination is not altered.
  ;
  ; If an invalid operation is detected, the modified data would be
  ; considered useless and the top two registers would be popped; also, if
  ; a memory location is specified as the destination, it would not be
  ; modified.
  ;
  ; -----------------------------------------------------------------------

    .386
    .model flat, stdcall  ; 32 bit memory model
    option casemap :none  ; case sensitive

    include ZLIB.inc

    .code

;#######################################################################

ZAtanh proc public USES ecx lpSrc1:DWORD, lpDest:DWORD, uID:DWORD
        
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      mov   ecx,uID
      and   ecx,Z_SINGLE or Z_DBL   ;check uID for operand size
      .if   ZERO?
            mov   ecx,10
            jmp   @F
      .elseif ecx == 4
            jmp   @F
      .elseif ecx == 8
            jmp   @F
      .endif
      xor   eax,eax                 ;error code
      ret                           ;no valid operand size

   @@:
      test  uID,SRC1_FPU            ;check if Src1 in on FPU
      .if   ZERO?                   ;Src1 not on FPU
            mov   eax,lpSrc1
            .if   ecx == 4
                  fld   dword ptr [eax+ecx]     ;imag1
                  fld   dword ptr [eax]         ;real1   imag1
            .elseif ecx == 8
                  fld   qword ptr [eax+ecx]
                  fld   qword ptr [eax]
            .elseif ecx == 10
                  fld   tbyte ptr [eax+ecx]
                  fld   tbyte ptr [eax]
            .endif
      .endif

; The initial addition functions will need cartesian coordinates

      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU     ;->cartesian
      .endif

;======
; 1 - z
;======
                        ;a    b
      fld   st(1)       ;B     a     b
      fchs              ;-B    a     b
      fld   st(1)       ;A     -B     a     b
      fchs              ;-A     -B    a     b
      fld1              ;1     -A     -B    a     b
      fadd  st(1),st    ;1     -A+1     -B    a     b

;======
; 1 + z
;======
      faddp st(3),st    ;-A+1     -B  a+1     b
                        ;A    B     a     b

;==============
; (1+z) / (1-z)
;==============
      fld   st          ;A    A     B  a     b
      fmul  st,st       ;A^2     A     B  a     b
      fld   st(2)       ;B  A^2     A     -B  a     b
      fmul  st,st       ;B^2     A^2     A     B    a     b
      fadd              ;A^2+B^2   A     B    a     b
      fstp  st(5)       ;A     B    a     b    A^2+B^2
      fld   st          ;A    A     B    a     b    A^2+B^2
      fmul  st,st(3)    ;aA    A     B    a     b    A^2+B^2
      fld   st(2)       ;B   aA    A     B    a     b    A^2+B^2
      fmul  st,st(5)    ;bB  aA    A     B    a     b    A^2+B^2
      fadd              ;bB+aA  A     B    a     b    A^2+B^2 
      fld   st(5)       ;A^2+B^2   bd+ac A     B    a     b    A^2+B^2
      fdiv              ;(bB+aA/A^2+B^2)   A     B    a     b    A^2+B^2
      fxch  st(4)       ;b   A     B    a    (bB+aA/A^2+B^2)     A^2+B^2
      fmul              ;b*A     B    a    (bB+aA/A^2+B^2)     A^2+B^2
      fxch  st(2)       ;a     B    b*A    (bB+aA/A^2+B^2)     A^2+B^2
      fmul              ;a*B    b*A    (bB+aA/A^2+B^2)     A^2+B^2
      fsub              ;b*A-a*B    (bB+aA/A^2+B^2)     A^2+B^2
      fdivrp st(2),st   ;(bB+aA/A^2+B^2)     (b*A-a*B/A^2+B^2)

;========================
; 1/2 * ln((1+z) / (1-z))
;========================
      invoke ZLn,0,0,SRC1_FPU or DEST_FPU
                        ;a    b
      fld1              ;1    a     b
      fchs              ;-1   a     b
      fxch              ;a    -1    b
      fscale            ;a/2   -1    b
      fxch  st(2)       ;b   -1    a/2
      fscale            ;b/2   -1   a/2
      fstp  st(1)       ;b/2   a/2
      fxch              ;a/2   b/2

      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            invoke ZCtoP,0,0,SRC1_FPU or DEST_FPU     ;->polar
      .endif

      fstsw ax                      ;get exception flags
      fwait
      shr   al,1                    ;test for invalid operation
      .if   CARRY?                  ;indicates invalid operation
            fstp  st                ;pop top register
            fstp  st                ;pop second register
            xor   eax,eax           ;error return code
            ret
      .endif

      test  uID,DEST_FPU            ;check destination of result
      .if   ZERO?                   ;to be stored in memory
            mov   eax,lpDest
            .if   ecx == 4
                  fstp  dword ptr [eax]         ;store REAL first
                  fstp  dword ptr [eax+ecx]     ;then store IMAG
            .elseif ecx == 8
                  fstp  qword ptr [eax]
                  fstp  qword ptr [eax+ecx]
            .else
                  fstp  tbyte ptr [eax]
                  fstp  tbyte ptr [eax+ecx]
            .endif
      .endif

      or    al,1                    ;to insure EAX!=0
      ret
    
ZAtanh endp

; #########################################################################

end
